import xtrack as xt
import xpart as xp

def test_optimize_with_radiation():
    
    env = xt.Environment()

    env.new('mb', xt.Bend, length=1, k0=1, h=1)
    env.new('mq', xt.Quadrupole, length=1, k1=1)
    env.new('ms', xt.Sextupole, length=1, k2=1)
    env.new('mo', xt.Octupole, length=1, k3=1)

    element_list = ['mb', 'mq', 'ms', 'mo']

    for element in element_list:
        line = env.new_line(components=[element])
        
        line.slice_thick_elements(
            slicing_strategies=[xt.Strategy(slicing=xt.Teapot(1))])
        
        line.build_tracker()
        line.configure_radiation('mean')
        
        part = xp.Particles(p0c=1e15, x=0.11)
        line.track(part)
        
        line.optimize_for_tracking()
        part_opt = xp.Particles(p0c=1e15, x=0.11)
        line.track(part_opt)
        
        assert part.x == part_opt.x
        assert part.y == part_opt.y
        assert part.px == part_opt.px
        assert part.py == part_opt.py
        assert part.zeta == part_opt.zeta
        assert part.pzeta == part_opt.pzeta

def test_optimize_with_delta_taper():
    
    env = xt.Environment()

    env.new('mb', xt.Bend, length=1, k0=1, h=1)
    env.new('mq', xt.Quadrupole, length=1, k1=1)
    env.new('ms', xt.Sextupole, length=1, k2=1)
    env.new('mo', xt.Octupole, length=1, k3=1)

    thin_classes = [xt.ThinSliceBendEntry, xt.ThinSliceBendExit, xt.ThinSliceBend, 
                    xt.ThinSliceQuadrupole, xt.ThinSliceSextupole, xt.ThinSliceOctupole]

    element_list = ['mb', 'mq', 'ms', 'mo']

    for element in element_list:
        line = env.new_line(components=[element])
        

        line.slice_thick_elements(
            slicing_strategies=[xt.Strategy(slicing=xt.Teapot(1))])

        for el in line.elements:
            if type(el) in thin_classes:
                el.delta_taper = 0.1
                print(el)
        
        line.build_tracker()
        line.configure_radiation('mean')
        
        part = xp.Particles(p0c=1e15, x=0.11)
        line.track(part)
        
        line.optimize_for_tracking()
        part_opt = xp.Particles(p0c=1e15, x=0.11)
        line.track(part_opt)
        
        assert part.x == part_opt.x
        assert part.y == part_opt.y
        assert part.px == part_opt.px
        assert part.py == part_opt.py
        assert part.zeta == part_opt.zeta
        assert part.pzeta == part_opt.pzeta



import xtrack as xt

# Create an environment
env = xt.Environment()

# Create a line with two quadrupoles and a marker
line = env.new_line(name='myline', components=[
    env.new('q0', xt.Quadrupole, length=2.0, at=10.),
    env.new('q1', xt.Quadrupole, length=2.0, at=20.),
    env.new('m0', xt.Marker, at=40.),
    ])

tt0 = line.get_table()
tt0.show(cols=['s_start', 's_center', 's_end'])
# is:
# name             s_start      s_center         s_end
# drift_1                0           4.5             9
# q0                     9            10            11
# drift_2               11            15            19
# q1                    19            20            21
# drift_3               21          30.5            40
# m0                    40            40            40
# _end_point            40            40            40

# Instantiate elements using the class directly
mysext =  xt.Sextupole(length=0.1, k2=0.2)
myaperture =  xt.LimitEllipse(a=0.01, b=0.02)

# Insert the element in the line and, contextually, define its name:
line.insert('s1', mysext, at=5., from_='q1')

# Alternatively, add the element to the environment and then do the insertion:
env.elements['ap1'] = myaperture
line.insert('ap1', at='q0@start')

tt = line.get_table()
tt.show(cols=['s_start', 's_center', 's_end'])
# is:
# name             s_start      s_center         s_end
# drift_1                0           4.5             9
# ap1                    9             9             9
# q0                     9            10            11
# drift_2               11            15            19
# q1                    19            20            21
# drift_3..0            21        22.975         24.95
# s1                 24.95            25         25.05
# drift_3..2         25.05        32.525            40
# m0                    40            40            40
# _end_point            40            40            40



import numpy as np
import matplotlib.pyplot as plt

import xtrack as xt
import xobjects as xo
import xpart as xp

import RF_Track as RFT

#############################################
##########    Input parameters    ###########
#############################################

# Bunch parameters
q0 = -1 # electrons
P0c = 50e6 # reference momentum, eV/c

# Solenoid parameters
B0 = 0.5 # T, on-axis field
R = 0.1 # m, aperture radius
Lsol = 1 # m, length

#############################################
#######  RF-Track's part starts here  #######
#############################################

# Setup the element
Sol = RFT.Solenoid(Lsol, B0, R)

# Setup the volume
Vsol = RFT.Volume()
Vsol.dt_mm = 1 # mm/c, integration step
Vsol.odeint_algorithm = 'rk2' # integration algorithm, e.g., 'rk2', 'rkf45', 'leapfrog' 
Vsol.set_s0(-0.75) # m, longitudinal start point
Vsol.set_s1( 0.75) # m, longitudinal end point
Vsol.add(Sol, 0, 0, 0, 'center');

#############################################
#######  RF-Track's part ends here    #######
#############################################

# Back to Xsuite

elements = {
    'sol': xt.RFT_Element(element=Vsol),
}

# Build the line
line = xt.Line(elements=elements, element_names=list(elements.keys()))
line.particle_ref = xt.Particles(p0c=P0c, mass0=xt.ELECTRON_MASS_EV, q0=q0)

# Define the beam using
sigma_x = 0.01 # m, sigma x
sigma_y = 0.01 # m, sigma y

## Build particle object on context
n_part = 10000

rng = np.random.default_rng(12345)

# track
## Choose a context
context = xo.ContextCpu()         # For CPU
# context = xo.ContextCupy()      # For CUDA GPUs
# context = xo.ContextPyopencl()  # For OpenCL GPUs

## Transfer lattice on context and compile tracking code
line.build_tracker(_context=context)

particles = xp.Particles(
    _context=context,
    x=rng.normal(scale=sigma_x, size=n_part),
    y=rng.normal(scale=sigma_y, size=n_part),
    p0c=P0c, mass0=xt.ELECTRON_MASS_EV, q0=q0)

#############################################
#########   Xsuite tracking    ##############
#############################################

print('tracking starts')
line.track(particles)
print('tracking ends')

#############################################
## Retrieve Twiss from RF-Track and plot  ###
#############################################

# Plot the transverse plane
plt.figure(1)
plt.scatter(particles.x*1e3, particles.y*1e3, s=10, facecolors='none', edgecolors='b')
plt.xlabel("$x$ [mm]")
plt.ylabel("$y$ [mm]")
plt.show()

# Plot the field along the axis
Z = np.linspace(-0.75, 0.75, 1001) # m
O = np.zeros(1001)
[E,B] = Vsol.get_field(O, O, Z*1e3, O) # X [mm], Y [mm], Z [mm], T [mm/c]
plt.figure(2)
plt.plot(Z.transpose(), B[:,2])
plt.xlabel("$S$ [m]")
plt.ylabel("$B_z$ [T]")
plt.show()



import pathlib

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()

@for_all_test_contexts
def test_radial_steering(test_context):

    line = xt.Line.from_json(test_data_folder /
                                    'psb_injection/line_and_particle.json')
    line.particle_ref = xp.Particles(
        mass0=(xp.PROTON_MASS_EV), q0=1, energy0=(2*160.+938.)*1.e6)
    line.build_tracker(_context=test_context)

    df_hz = 180 # Frequency trim

    tw = line.twiss()

    h_rf = 1
    f_rf = h_rf/tw.T_rev0

    beta0 = line.particle_ref.beta0[0]

    # T_ref = h_rf/f_rf
    # dt = h_rf/(f_rf + df_hz) - h_rf/f_rf = h_rf/f_rf (1/(1+df_hz/f_rf) - 1)
    #                                       ~= h_rf/f_rf * (1 - df_hz/f_rf -1)
    #                                       = -h_rf/(f_rf^2) * df_hz
    #                                       = -T_rev / f_rf * df_hz
    # dzeta = -beta0 * clight * dt = circumference * df_hz / f_rf

    dzeta = tw.circumference * df_hz / f_rf

    line.unfreeze()
    line.append_element(element=xt.ZetaShift(dzeta=dzeta), name='zeta_shift')
    line.build_tracker()

    tw_6d_offmom = line.twiss()

    print(f'delta closed orbit: {tw_6d_offmom.delta[0]:.3e}')
    # prints: delta closed orbit: -1.288e-03

    # Checks
    eta = tw.slip_factor
    f0 = 1/tw.T_rev0
    delta_trim = -1/h_rf/eta/f0*df_hz

    # Use 4d twiss on machine without zeta shift
    line['zeta_shift'].dzeta = 0
    tw_on_mom = line.twiss(delta0=0, method='4d')
    tw_off_mom = line.twiss(delta0=delta_trim, method='4d')
    dzeta_from_twiss = (tw_off_mom['zeta'][-1] - tw_off_mom['zeta'][0])

    print(f'delta_trim: {delta_trim}')
    print(f'tw_6d_offmom.delta[0]: {tw_6d_offmom.delta[0]}')
    print(f'tw_6d_offmom.ptau[0]: {tw_6d_offmom.ptau[0]}')

    xo.assert_allclose(delta_trim, tw_6d_offmom.delta[0], rtol=1e-3, atol=0)
    xo.assert_allclose(dzeta, dzeta_from_twiss, rtol=1e-3, atol=0)


import xtrack as xt
import xobjects as xo
import numpy as np

env = xt.Environment()

line = env.new_line(
    components=[
        env.new('q0', 'Quadrupole', length=2.0, at=20.0),
        env.new('ql', 'Quadrupole', length=2.0, at=10.0),
        env.new('qr', 'Quadrupole', length=2.0, at=30),
        env.new('end', 'Marker', at=50.),
    ])

ln_insert = env.new_line(
    components=[
        env.new('s1', 'Sextupole', length=0.1),
        env.new('s2', 's1', anchor='start', at=0.3, from_='s1@end'),
        env.new('s3', 's1', anchor='start', at=0.3, from_='s2@end')
    ])

line.insert(ln_insert, anchor='start', at=1, from_='q0@end')

tt = line.get_table()
tt.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt.name == np.array(
    ['drift_1', 'ql', 'drift_2', 'q0', 'drift_3..0', 's1', 'drift_5',
       's2', 'drift_6', 's3', 'drift_3..6', 'qr', 'drift_4', 'end',
       '_end_point']))
xo.assert_allclose(tt.s_center, np.array(
    [ 4.5 , 10.  , 15.  , 20.  , 21.5 , 22.05, 22.25, 22.45, 22.65,
       22.85, 25.95, 30.  , 40.5 , 50.  , 50.  ]),
    rtol=0., atol=1e-14)


import numpy as np
import xtrack as xt

# Load a machine model
line = xt.Line.from_json(
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.cycle('ip1', inplace=True)
line.twiss_default['method'] = '4d'

# Select circuits with appropriate weights
vary=[xt.VaryList(['kqs.a23b1', 'kqs.a67b1'], step=5e-5),
      xt.VaryList(['kqs.l4b1', 'kqs.l8b1','kqs.r3b1', 'kqs.r7b1'],
                  weight=2, step=5e-5)]

# Match c_minus_re.b1
c_min_match = 1e-4
opt_re = line.match_knob(knob_name='c_minus_re.b1',
    knob_value_start=0, knob_value_end=c_min_match,
    run=False,
    vary=vary,
    targets=[
        xt.Target('c_minus_re_0', value=c_min_match, tol=1e-8),
        xt.Target('c_minus_im_0', value=0,           tol=1e-8),
    ])
opt_re.solve()
opt_re.generate_knob()

# Match c_minus_im.b1
opt_im = line.match_knob(knob_name='c_minus_im.b1',
    knob_value_start=0, knob_value_end=c_min_match,
    run=False,
    vary=vary,
    targets=[
        xt.Target('c_minus_re_0', value=0,           tol=1e-8),
        xt.Target('c_minus_im_0', value=c_min_match, tol=1e-8),
    ])
opt_im.solve()
opt_im.generate_knob()

# Test the knob
line['c_minus_re.b1'] = 1e-3
tw = line.twiss4d()
tw.c_minus_re_0 # is 0.00099998
tw.c_minus_im_0 # is 4.05e-09

line['c_minus_re.b1'] = 0
line['c_minus_im.b1'] = 1e-3
tw = line.twiss4d()
tw.c_minus_re_0 # is 2.16e-8
tw.c_minus_im_0 # is 0.001000003



import xtrack as xt

# Load a line and build a tracker
collider = xt.Environment.from_json(
    '../../test_data/hllhc15_thick/hllhc15_collider_thick.json')
collider.build_trackers()

tw0 = collider.twiss(method='4d')

opt = collider.match(
    lines=['lhcb1', 'lhcb2'],
    start=['mq.30l8.b1', 'mq.31l8.b2'],
    end=['mq.23l8.b1', 'mq.24l8.b2'],
    init=tw0,
    vary=xt.VaryList(['acbv30.l8b1', 'acbv28.l8b1', 'acbv26.l8b1', 'acbv24.l8b1',
                      'acbv31.l8b2', 'acbv29.l8b2', 'acbv27.l8b2', 'acbv25.l8b2'],
                    step=1e-10, limits=[-1e-3, 1e-3]),
    targets = [
        xt.TargetSet(y=3e-3, py=0, at='mb.b28l8.b1', line='lhcb1'),
        xt.TargetSet(y=-3e-3, py=0, at='mb.b28l8.b2', line='lhcb2'),
        xt.TargetSet(y=0, py=0, at=xt.END, line='lhcb1'),
        xt.TargetSet(y=0, py=0, at=xt.END, line='lhcb2'),
    ])

opt.target_status()

#!end-doc-part
import matplotlib.pyplot as plt
import numpy as np
plt.close('all')

line = collider['lhcb1']

opt.tag(tag='matched')
opt.reload(0)
tw_before = line.twiss(method='4d')
opt.reload(tag='matched')
tw = line.twiss(method='4d')

fig = plt.figure(1, figsize=(6.4*1.2, 4.8*0.8))
ax = fig.add_subplot(111)
ax.plot(tw.s, tw.y*1000, label='y')

for nn in ['mcbv.30l8.b1', 'mcbv.28l8.b1', 'mcbv.26l8.b1', 'mcbv.24l8.b1']:
    ax.axvline(x=line.get_s_position(nn), color='k', linestyle='--', alpha=0.5)
    ax.text(line.get_s_position(nn), 10, nn, rotation=90,
            horizontalalignment='left', verticalalignment='top')

ax.axvline(x=line.get_s_position('mb.b28l8.b1'), color='r', linestyle='--', alpha=0.5)
ax.text(line.get_s_position('mb.b28l8.b1'), 10, 'mb.b28l8.b1', rotation=90,
        horizontalalignment='left', verticalalignment='top')

ax.axvline(x=line.get_s_position('mq.30l8.b1'), color='g', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mq.23l8.b1'), color='g', linestyle='--', alpha=0.5)
ax.text(line.get_s_position('mq.30l8.b1'), 10, 'mq.30l8.b1', rotation=90,
        horizontalalignment='right', verticalalignment='top')
ax.text(line.get_s_position('mq.23l8.b1'), 10, 'mq.23l8.b1', rotation=90,
        horizontalalignment='right', verticalalignment='top')

ax.set_xlim(line.get_s_position('mq.30l8.b1') - 10,
        line.get_s_position('mq.23l8.b1') + 10)
ax.set_xlabel('s [m]')
ax.set_ylabel('y [mm]')
ax.set_ylim(-0.5, 10)
plt.subplots_adjust(bottom=.152, top=.9, left=.1, right=.95)

assert np.isclose(tw['y', 'mb.b28l8.b1'], 3e-3, atol=1e-4)
assert np.isclose(tw['py', 'mb.b28l8.b1'], 0, atol=1e-6)
assert np.isclose(tw['y', 'mq.23l8.b1'], tw_before['y', 'mq.23l8.b1'], atol=1e-6)
assert np.isclose(tw['py', 'mq.23l8.b1'], tw_before['py', 'mq.23l8.b1'], atol=1e-7)
assert np.isclose(tw['y', 'mq.33l8.b1'], tw_before['y', 'mq.33l8.b1'], atol=1e-6)
assert np.isclose(tw['py', 'mq.33l8.b1'], tw_before['py', 'mq.33l8.b1'], atol=1e-7)


line = collider['lhcb2']

opt.tag(tag='matched')
opt.reload(0)
tw_before = line.twiss(method='4d')
opt.reload(tag='matched')
tw = line.twiss(method='4d')

fig = plt.figure(2, figsize=(6.4*1.2, 4.8*0.8))
ax = fig.add_subplot(111)
ax.plot(tw.s, tw.y*1000, label='y')

for nn in ['mcbv.31l8.b2', 'mcbv.29l8.b2', 'mcbv.27l8.b2', 'mcbv.25l8.b2']:
    ax.axvline(x=tw['s', nn], color='k', linestyle='--', alpha=0.5)
    ax.text(tw['s', nn], -9.8, nn, rotation=90,
            horizontalalignment='left', verticalalignment='bottom')

ax.axvline(x=tw['s', 'mb.b28l8.b2'], color='r', linestyle='--', alpha=0.5)
ax.text(tw['s', 'mb.b28l8.b2'], -9.8, 'mb.b28l8.b2', rotation=90,
        horizontalalignment='left', verticalalignment='bottom')

ax.axvline(x=tw['s', 'mq.31l8.b2'], color='g', linestyle='--', alpha=0.5)
ax.axvline(x=tw['s', 'mq.24l8.b2'], color='g', linestyle='--', alpha=0.5)
ax.text(tw['s', 'mq.31l8.b2'], -9.8, 'mq.31l8.b2', rotation=90,
        horizontalalignment='right', verticalalignment='bottom')
ax.text(tw['s', 'mq.24l8.b2'], -9.8, 'mq.24l8.b2', rotation=90,
        horizontalalignment='right', verticalalignment='bottom')

ax.set_xlim(tw['s', 'mq.31l8.b2'] - 10, tw['s', 'mq.24l8.b2'] + 10)
ax.set_xlabel('s [m]')
ax.set_ylabel('y [mm]')
ax.set_ylim(-10, 0.5)
plt.subplots_adjust(bottom=.152, top=.9, left=.1, right=.95)

assert np.isclose(tw['y', 'mb.b28l8.b2'], -3e-3, atol=1e-4)
assert np.isclose(tw['py', 'mb.b28l8.b2'], 0, atol=1e-6)
assert np.isclose(tw['y', 'mq.24l8.b2'], tw_before['y', 'mq.23l8.b2'], atol=1e-6)
assert np.isclose(tw['py', 'mq.24l8.b2'], tw_before['py', 'mq.23l8.b2'], atol=1e-7)
assert np.isclose(tw['y', 'mq.33l8.b2'], tw_before['y', 'mq.33l8.b2'], atol=1e-6)
assert np.isclose(tw['py', 'mq.33l8.b2'], tw_before['py', 'mq.33l8.b2'], atol=1e-7)

plt.show()


import xtrack as xt
import xpart as xp

from cpymad.madx import Madx

mad1=Madx()
mad1.call('../../test_data/hllhc15_thick/lhc.seq')
mad1.call('../../test_data/hllhc15_thick/hllhc_sequence.madx')
mad1.input('beam, sequence=lhcb1, particle=proton, energy=7000;')
mad1.use('lhcb1')
mad1.call("../../test_data/hllhc15_thick/opt_round_150_1500.madx")
mad1.twiss()

mad4=Madx()
mad4.input('mylhcbeam=4')
mad4.call('../../test_data/hllhc15_thick/lhcb4.seq')
mad4.call('../../test_data/hllhc15_thick/hllhc_sequence.madx')
mad4.input('beam, sequence=lhcb2, particle=proton, energy=7000;')
mad4.use('lhcb2')
mad4.call("../../test_data/hllhc15_thick/opt_round_150_1500.madx")
mad4.twiss()


line1=xt.Line.from_madx_sequence(mad1.sequence.lhcb1,
                                 allow_thick=True,
                                 deferred_expressions=True,
                                 replace_in_expr={'bv_aux':'bvaux_b1'})
line1.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, p0c=7000e9)

line4=xt.Line.from_madx_sequence(mad4.sequence.lhcb2,
                                 allow_thick=True,
                                 deferred_expressions=True,
                                 replace_in_expr={'bv_aux':'bvaux_b2'})
line4.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, p0c=7000e9)

# Remove solenoids (cannot backtwiss for now)
for ll in [line1, line4]:
    tt = ll.get_table()
    for nn in tt.rows[tt.element_type=='Solenoid'].name:
        ee_elen = ll[nn].length
        ll.element_dict[nn] = xt.Drift(length=ee_elen)

collider = xt.Environment(lines={'lhcb1':line1,'lhcb2':line4})
collider.lhcb1.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, p0c=7000e9)
collider.lhcb2.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, p0c=7000e9)

collider.lhcb1.twiss_default['method'] = '4d'
collider.lhcb2.twiss_default['method'] = '4d'
collider.lhcb2.twiss_default['reverse'] = True

collider.to_json('hllhc15_collider_thick.json')


import pathlib
from cpymad.madx import Madx
import numpy as np
import xtrack as xt
import xobjects as xo

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()

def test_h6_sps_beamline():

    '''
    This checks the parsing of a sequence defined in madx with refer=exit
    and the TwissTable.get_R_matrix_table() method
    '''

    lattice_path = test_data_folder / 'h6_experimental_line/h6-fm.str'
    sequence_path = test_data_folder / 'h6_experimental_line/h6fm04.seq'

    mad = Madx()
    mad.call(str(lattice_path))
    mad.call(str(sequence_path))

    mad.input('''
    beam, particle=proton, sequence=H6, PC=120.0,
        ex = 2e-07,
        ey = 5e-08;

    use, sequence=H6;
    twiss, chrom=true, rmatrix=true, betx=10, alfx=0, bety=10, alfy=0;
    ''')
    tw_mad = xt.Table(mad.table.twiss)

    env = xt.load_madx_lattice(sequence_path)
    env.vars.load_madx(lattice_path)
    line = env['h6']
    line.particle_ref = xt.Particles(p0c=120e9, mass0=xt.PROTON_MASS_EV)
    tt = line.get_table(attr=True)
    line.configure_bend_model(edge='full')

    tw = line.twiss(betx=10, bety=10)

    check_at = [
    't4..centre',
    'vxsv.x0410104',
    'begin.vac',
    'xwca.x0410404',
    'xsci.x0410475',
    'xemc.x0410476',
    'xdwc.x0410488',
    'h6a',
    'h6b',
    'h6c']

    tw_check = tw.rows[check_at]
    tw_mad_check = tw_mad.rows[[nn+':1' for nn in check_at]]

    xo.assert_allclose(tw_check.betx, tw_mad_check.betx, rtol=2e-5, atol=0)
    xo.assert_allclose(tw_check.bety, tw_mad_check.bety, rtol=2e-5, atol=0)
    xo.assert_allclose(tw_check.dx, tw_mad_check.dx, rtol=2e-5, atol=1e-4)
    xo.assert_allclose(tw_check.dy, tw_mad_check.dy, rtol=2e-5, atol=1e-4)

    trm = tw.get_R_matrix_table()
    trm_check = trm.rows[check_at]

    for ii in range(6):
        for jj in range(6):
            rterm_mad = tw_mad_check[f're{ii+1}{jj+1}']
            rterm_xs = trm_check[f'r{ii+1}{jj+1}']
            atol=1e-4*np.max(np.abs(rterm_mad))
            if atol<1e-14:
                atol=1e-14
            xo.assert_allclose(rterm_xs, rterm_mad, rtol=2e-5, atol=atol)


import json
import numpy as np
import xtrack as xt


case_name = 'fcc-ee'
filename = 'line_no_radiation.json'

with open(filename, 'r') as f:
    line = xt.Line.from_dict(json.load(f))

#line['qc1l1.1..1'].knl[0] += 1e-5
#line['qc1l1.1..1'].ksl[0] += 1e-6/5
#line['qc1l1.1..1'].ksl[1] = 1e-4

line.build_tracker()
line.config.XTRACK_GLOBAL_XY_LIMIT = 20e-2

for ee in line.elements:
    if ee.__class__.__name__.startswith('Cavity'):
        ee.voltage = 0

# Initial twiss (no radiation)
line.configure_radiation(model=None)
tw_no_rad = line.twiss(method='4d', freeze_longitudinal=True)

# Enable radiation
line.configure_radiation(model='mean')

p = tw_no_rad.particle_on_co.copy()

line.track(p, turn_by_turn_monitor='ONE_TURN_EBE')
mon = line.record_last_track
mon.x[mon.state<1] = np.nan
mon.y[mon.state<1] = np.nan
mon.delta[mon.state<1] = np.nan

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
ax1 = plt.subplot(211)
plt.plot(mon.s.T, mon.x.T, label='x')
plt.plot(mon.s.T, mon.y.T, label='y')
plt.legend(loc='lower left')
plt.ylabel('x, y [m]')
ax2 = plt.subplot(212, sharex=ax1)
plt.plot(mon.s.T, mon.delta.T)
plt.xlabel('s [m]')
plt.ylabel(r'$\delta$')

plt.show()


from typing import List

import numpy as np
from scipy.special import factorial

import xobjects as xo
from xtrack.base_element import BeamElement
from xtrack.general import _pkg_root
from ..internal_record import RecordIndex
from ..random import RandomUniformAccurate, RandomExponential

DEFAULT_MULTIPOLE_ORDER = 5


_INDEX_TO_MODEL_CURVED = {
    0: 'adaptive',
    1: 'full',
    2: 'bend-kick-bend',
    3: 'rot-kick-rot',
    4: 'mat-kick-mat',
    5: 'drift-kick-drift-exact',
    6: 'drift-kick-drift-expanded',
}
_MODEL_TO_INDEX_CURVED = {k: v for v, k in _INDEX_TO_MODEL_CURVED.items()} | {'expanded': 4}

_INDEX_TO_INTEGRATOR = {
    0: 'adaptive',
    1: 'teapot',
    2: 'yoshida4',
    3: 'uniform',
}
_INTEGRATOR_TO_INDEX = {k: v for v, k in _INDEX_TO_INTEGRATOR.items()}

_INDEX_TO_EDGE_MODEL = {
   -1: 'suppressed',
    0: 'linear',
    1: 'full',
    2: 'dipole-only',
}
_EDGE_MODEL_TO_INDEX = {k: v for v, k in _INDEX_TO_EDGE_MODEL.items()}

_INDEX_TO_MODEL_STRAIGHT = _INDEX_TO_MODEL_CURVED.copy()
_INDEX_TO_MODEL_STRAIGHT.pop(2)
_INDEX_TO_MODEL_STRAIGHT.pop(3)
_MODEL_TO_INDEX_STRAIGHT = {k: v for v, k in _INDEX_TO_MODEL_STRAIGHT.items()}

_NOEXPR_FIELDS = {'model', 'integrator', 'edge_entry_model', 'edge_exit_model'}

COMMON_MAGNET_SOURCES = [
    _pkg_root.joinpath('headers/synrad_spectrum.h'),
    _pkg_root.joinpath('beam_elements/elements_src/track_yrotation.h'),
    _pkg_root.joinpath('beam_elements/elements_src/track_wedge.h'),
    _pkg_root.joinpath('beam_elements/elements_src/track_dipole_fringe.h'),
    _pkg_root.joinpath('beam_elements/elements_src/track_dipole_edge_linear.h'),
    _pkg_root.joinpath('beam_elements/elements_src/track_mult_fringe.h'),
    _pkg_root.joinpath('beam_elements/elements_src/track_magnet_edge.h'),
    _pkg_root.joinpath('beam_elements/elements_src/track_magnet_drift.h'),
    _pkg_root.joinpath('beam_elements/elements_src/track_magnet_kick.h'),
    _pkg_root.joinpath('beam_elements/elements_src/track_magnet_radiation.h'),
    _pkg_root.joinpath('beam_elements/elements_src/track_magnet.h'),
]


class SynchrotronRadiationRecord(xo.HybridClass):
    _xofields = {
        '_index': RecordIndex,
        'photon_energy': xo.Float64[:],
        'at_element': xo.Int64[:],
        'at_turn': xo.Int64[:],
        'particle_id': xo.Int64[:],
        'particle_delta': xo.Float64[:]
    }


class MagnetDrift(BeamElement):
    """A drift slice in a magnet kick splitting. Mostly used for testing purposes.

    See ``Magnet`` for the description of the parameters.
    """
    isthick = True
    has_backtrack = True

    _xofields = {
        'length': xo.Float64,
        'k0': xo.Float64,
        'k1': xo.Float64,
        'h': xo.Float64,
        'drift_model': xo.Int64,
    }

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/track_magnet_drift.h'),
        _pkg_root.joinpath('beam_elements/elements_src/magnet_drift.h'),
    ]


class MagnetKick(BeamElement):
    """A thin kick in a magnet kick splitting. Mostly used for testing purposes.

    See ``Magnet`` for the description of the parameters.
    """
    isthick = False
    has_backtrack = False

    _xofields = {
        'length': xo.Float64,
        'order': xo.Int64,
        'inv_factorial_order': xo.Float64,
        'knl': xo.Float64[:],
        'ksl': xo.Float64[:],
        'factor_knl_ksl': xo.Float64,
        'kick_weight': xo.Float64,
        'k0': xo.Float64,
        'k1': xo.Float64,
        'k2': xo.Float64,
        'k3': xo.Float64,
        'k0s': xo.Float64,
        'k1s': xo.Float64,
        'k2s': xo.Float64,
        'k3s': xo.Float64,
        'h': xo.Float64,
    }

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/track_magnet_kick.h'),
        _pkg_root.joinpath('beam_elements/elements_src/magnet_kick.h'),
    ]


class Magnet(BeamElement):
    """General transverse field magnet with curvature and fringe fields.

    A beam element representing a magnet with transverse fields, curvature, and
    edge and fringe-field effects. Optional ``integrator`` and ``model``
    parameters can be used to specify the integration scheme and drift model to
    be used in the kick-splitting scheme. Default value is ``adaptive`` for
    both, which aims to provide best results in the general case (``rot-kick-rot``
    using the polar/exact, drift depending on h, for the model, and ``yoshida4``
    for the integration scheme).

    Parameters
    ----------
    length : float, optional
        Length of the element in meters along the reference trajectory.
    k0 : float, optional
        Strength of the horizontal dipolar component in units of m^-1.
    k1 : float, optional
        Strength of the horizontal quadrupolar component in units of m^-2.
    k2 : float, optional
        Strength of the horizontal sextupolar component in units of m^-3.
    k3 : float, optional
        Strength of the horizontal octupolar component in units of m^-4.
    k0s : float, optional
        Strength of the skew dipolar component in units of m^-1.
    k1s : float, optional
        Strength of the skew quadrupolar component in units of m^-2.
    k2s : float, optional
        Strength of the skew sextupolar component in units of m^-3.
    k3s : float, optional
        Strength of the skew octupolar component in units of m^-4.
    h : float, optional
        Curvature of the reference trajectory in units of m^-1 (= 1 / radius).
        Will imply the value of ``k0`` if ``k0_from_h`` is set.
    k0_from_h : bool, optional
        If true, the value of ``k0`` will be pinned to the value of ``h``.
    order : int, optional
        Maximum order of multipole expansion for this magnet. Defaults to 5.
    knl : list of floats, optional
        Normal multipole integrated strengths. If not provided, defaults to zeroes.
    ksl : list of floats, optional
        Skew multipole integrated strengths. If not provided, defaults to zeroes.
    model : str, optional
        Drift model to be used in the kick-splitting scheme. The options are:

            - ``adaptive``: default option, same as ``rot-kick-rot``.
            - ``full``: kept for backward compatibility, same as ``rot-kick-rot``.
            - ``bend-kick-bend``: use a thick (curved, if ``h`` non-zero) exact
                bend map for ``k0``, ``h``, and handle the other strengths in
                the kicks.
            - ``rot-kick-rot``: use an exact drift map (polar, if ``h`` non-zero)
                and handle all strengths in the kicks.
            - ``mat-kick-mat``: use an expanded combined-function magnet map
                for ``k0``, ``k1``, ``h``, and handle the other strengths in
                the kicks.
            - ``drift-kick-drift-exact``: use an exact drift map with no curvature,
                and handle all strengths in the kicks.
            - ``drift-kick-drift-expanded``: use an expanded drift map with no
                curvature, and handle all strengths in the kicks.

        These will not be applied if the length is zero.
    integrator : str, optional
        Integration scheme to be used. The options are:

            - ``adaptive``: default option, same as ``yoshida4``.
            - ``teapot``: use the Teapot integration scheme.
            - ``yoshida4``: use the Yoshida 4 integration scheme. The number of
                kicks will be implicitly rounded up to the nearest multiple of 7,
                as required by the scheme.
            - ``uniform``: slice uniformly.

        The integration scheme setting will be ignored if the length is zero, or
        if the strength and the curvature settings imply no need for applying
        thin kicks.
    num_multipole_kicks : int, optional
        The number of kicks to be used in thin kick splitting. If zero, and if
        the model selection implies that there are kicks that need to be
        performed, the value will be guessed according to a heuristic: one kick
        in the middle for straight magnets, or ~2 kicks/mrad otherwise.
    edge_entry_active : bool, optional
        Whether to include the edge effect at entry. Enabled by default.
    edge_exit_active : bool, optional
        Whether to include the edge effect at exit. Enabled by default.
    edge_entry_model : str, optional
        Edge model at magnet entry. The options are:

            - ``linear``: use a linear model for the edge.
            - ``full``: include all multipolar terms.
            - ``dipole-only``: ``full`` but includes only the dipolar terms.
            - ``suppressed``: ignore the edge effect.
    edge_exit_model : str, optional
        Edge model at magnet exit. See ``edge_entry_model`` for the options.
    edge_entry_angle : float, optional
        The angle of the entry edge in radians. Default is 0.
    edge_exit_angle : float, optional
        Same as `edge_entry_angle`, but for the exit.
    edge_entry_angle_fdown : float, optional
        Term added to the entry angle only for the ``linear`` mode and only in
        the vertical plane to account for non-zero angle in the closed orbit
        when entering the fringe field (feed down effect). Default is 0.
    edge_exit_angle_fdown : float, optional
        Same as ``edge_entry_angle_fdown``, but for the exit. Default is 0.
    edge_entry_fint: float, optional
        Fringe integral value at entry. Default is 0.
    edge_exit_fint : float, optional
        Same as ``edge_entry_fint``, but for the exit. Default is 0.
    edge_entry_hgap : float, optional
        Equivalent gap at entry in meters. Default is 0.
    edge_exit_hgap : float, optional
        Same as ``edge_entry_hgap``, but for the exit.
    radiation_flag : int, optional
        Flag indicating if synchrotron radiation effects are enabled.
        If zero, no radiation effects are simulated; if 1, the ``mean``
        model is used; if 2, the ``quantum`` model is used and the
        emitted photons are stored in the internal radiation record.
    delta_taper : float, optional
        A value added to delta for the purposes of tapering. Default is 0.
    """
    isthick = True
    has_backtrack = True

    _xofields = {
        'length': xo.Float64,
        'order': xo.Int64,
        'inv_factorial_order': xo.Float64,
        'num_multipole_kicks': xo.Int64,
        'knl': xo.Float64[:],
        'ksl': xo.Float64[:],
        'k0': xo.Float64,
        'k1': xo.Float64,
        'k2': xo.Float64,
        'k3': xo.Float64,
        'k0s': xo.Float64,
        'k1s': xo.Float64,
        'k2s': xo.Float64,
        'k3s': xo.Float64,
        'angle': xo.Float64,
        'h': xo.Float64,
        'k0_from_h': xo.UInt64,
        'edge_entry_active': xo.Field(xo.Int64, default=1),
        'edge_exit_active': xo.Field(xo.Int64, default=1),
        'edge_entry_model': xo.Int64,
        'edge_exit_model': xo.Int64,
        'edge_entry_angle': xo.Float64,
        'edge_exit_angle': xo.Float64,
        'edge_entry_angle_fdown': xo.Float64,
        'edge_exit_angle_fdown': xo.Float64,
        'edge_entry_fint': xo.Float64,
        'edge_exit_fint': xo.Float64,
        'edge_entry_hgap': xo.Float64,
        'edge_exit_hgap': xo.Float64,
        'model': xo.Int64,
        'integrator': xo.Int64,
        'radiation_flag': xo.Int64,
        'delta_taper': xo.Float64,
    }

    _rename = {
        'order': '_order',
        'model': '_model',
        'edge_entry_model': '_edge_entry_model',
        'edge_exit_model': '_edge_exit_model',
        'k0': '_k0',
        'k0_from_h': '_k0_from_h',
        'angle': '_angle',
        'length': '_length',
        'h': '_h',
        'integrator': '_integrator',
    }

    _extra_c_sources = [
        *COMMON_MAGNET_SOURCES,
        _pkg_root.joinpath('beam_elements/elements_src/magnet.h'),
    ]

    _depends_on = [RandomUniformAccurate, RandomExponential]

    _internal_record_class = SynchrotronRadiationRecord

    def __init__(self, order=None, knl: List[float]=None, ksl: List[float]=None, **kwargs):
        if '_xobject' in kwargs.keys() and kwargs['_xobject'] is not None:
            self.xoinitialize(**kwargs)
            return

        order = order or DEFAULT_MULTIPOLE_ORDER
        multipolar_kwargs = _prepare_multipolar_params(order, knl=knl, ksl=ksl)
        kwargs.update(multipolar_kwargs)

        model = kwargs.pop('model', None)
        integrator = kwargs.pop('integrator', None)
        edge_entry_model = kwargs.pop('edge_entry_model', None)
        edge_exit_model = kwargs.pop('edge_exit_model', None)

        self.xoinitialize(**kwargs)

        # Calculate length and h in the event length_straight and/or angle given
        self.set_bend_params(
            kwargs.get('length'),
            kwargs.get('h'),
            kwargs.get('angle'),
        )

        if self.k0_from_h:
            self.k0 = self.h

        # Trigger properties
        if model is not None:
            self.model = model

        if integrator is not None:
            self.integrator = integrator

        if edge_entry_model is not None:
            self.edge_entry_model = edge_entry_model

        if edge_exit_model is not None:
            self.edge_exit_model = edge_exit_model

    def set_bend_params(self, length=None, h=None, angle=None):
        length, h, angle = self.compute_bend_params(
            length, h, angle,
        )

        # None becomes NaN in numpy buffers
        if length is not None:
            self._length = length
        if h is not None:
            self._h = h
        if angle is not None:
            self._angle = angle

        if self.k0_from_h:
            self._k0 = self.h

    @staticmethod
    def compute_bend_params(length=None, h=None, angle=None):
        if not length:
            # If no length, then we cannot meaningfully calculate anything
            return length, h, angle

        if angle is not None:
            computed_h = angle / length

            if h is not None and not np.isclose(h, computed_h, rtol=0, atol=1e-13):
                raise ValueError('Given `h` and `angle` are inconsistent!')

            h = h or computed_h
            return length, h, angle

        if h is not None:
            computed_angle = h * length

            if angle is not None and not np.isclose(angle, computed_angle, rtol=0, atol=1e-13):
                raise ValueError('Given `h` and `angle` are inconsistent!')

            angle = angle or computed_angle
            return length, h, angle

        # Both `h` and `angle` are None
        return length, h, angle

    @property
    def k0(self):
        return self._k0

    @k0.setter
    def k0(self, value):
        if self.k0_from_h and not np.isclose(value, self.h, atol=1e-13):
            self.k0_from_h = False
        self._k0 = value

    @property
    def k0_from_h(self):
        return bool(self._k0_from_h)

    @k0_from_h.setter
    def k0_from_h(self, value):
        if value:
            self._k0 = self.h
        self._k0_from_h = value

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, value):
        self.set_bend_params(length=value, angle=self.angle)

    @property
    def h(self):
        return self._h

    @h.setter
    def h(self, value):
        self.set_bend_params(length=self.length, h=value)

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, value):
        self.set_bend_params(length=self.length, angle=value)

    @property
    def order(self):
        return self._order

    @order.setter
    def order(self, value):
        self._order = value
        self.inv_factorial_order = 1.0 / factorial(value, exact=True)

    @property
    def model(self):
        return _INDEX_TO_MODEL_CURVED[self._model]

    @model.setter
    def model(self, value):
        try:
            self._model = _MODEL_TO_INDEX_CURVED[value]
        except KeyError:
            raise ValueError(f'Invalid model: {value}')

    @property
    def integrator(self):
        return _INDEX_TO_INTEGRATOR[self._integrator]

    @integrator.setter
    def integrator(self, value):
        try:
            self._integrator = _INTEGRATOR_TO_INDEX[value]
        except KeyError:
            raise ValueError(f'Invalid integrator: {value}')

    @property
    def edge_entry_model(self):
        return _INDEX_TO_EDGE_MODEL[self._edge_entry_model]

    @edge_entry_model.setter
    def edge_entry_model(self, value):
        try:
            self._edge_entry_model = _EDGE_MODEL_TO_INDEX[value]
        except KeyError:
            raise ValueError(f'Invalid edge model: {value}')

    @property
    def edge_exit_model(self):
        return _INDEX_TO_EDGE_MODEL[self._edge_exit_model]

    @edge_exit_model.setter
    def edge_exit_model(self, value):
        try:
            self._edge_exit_model = _EDGE_MODEL_TO_INDEX[value]
        except KeyError:
            raise ValueError(f'Invalid edge model: {value}')


class MagnetEdge(BeamElement):
    """Beam element modeling a magnet edge. Mostly used for testing purposes.

    Parameters
    ----------
    model : str
        Model to be used for the edge. See ``Magnet.edge_entry_model`` and
        ``Magnet.edge_exit_model`` for the options.
    is_exit : bool
        If False, the edge is the entrance edge. If True, the edge is an exit edge.
    kn : list of floats
        List of normal multipolar strengths. If not provided, will be filled
        with zeros according to ``k_order``.
    ks : list of floats
        List of skew multipolar strengths. If not provided, will be filled
        with zeros according to ``k_order``.
    k_order : int
        Order of kn and ks. If not provided, will either be inferred from kn
        and/or ks or set to -1.
    knl : list of floats
        List of integrated normal strengths. If not provided, will be filled
        with zeros according to ``kl_order``.
    ksl : list of floats
        List of integrated skew strengths. If not provided, will be filled
        with zeros according to ``kl_order``.
    kl_order : int
        Order of knl and ksl. If not provided, will either be inferred from
        knl and/or ksl or set to -1.
    length : float
        Length of the magnet. Only necessary if integrated strengths are given.
    half_gap : float
        Equivalent gap in m.
    face_angle : float
        Face angle in rad.
    face_angle_feed_down : float
        Term added to ``face_angle`` only for the linear mode and only in the
        vertical plane to account for non-zero angle in the closed orbit when
        entering the fringe field (feed down effect).
    fringe_integral : float
        Fringe integral.
    """
    isthick = True
    has_backtrack = True

    _xofields = {
        'model': xo.Int64,
        'is_exit': xo.Int64,
        'kn': xo.Float64[:],
        'ks': xo.Float64[:],
        'k_order': xo.Field(xo.Int64, default=-1),
        'knl': xo.Float64[:],
        'ksl': xo.Float64[:],
        'kl_order': xo.Field(xo.Int64, default=-1),
        'length': xo.Float64,
        'half_gap': xo.Float64,
        'face_angle': xo.Float64,
        'face_angle_feed_down': xo.Float64,
        'fringe_integral': xo.Float64,
    }

    _rename = {
        'model': '_model',
    }

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/track_yrotation.h'),
        _pkg_root.joinpath('beam_elements/elements_src/track_wedge.h'),
        _pkg_root.joinpath('beam_elements/elements_src/track_dipole_fringe.h'),
        _pkg_root.joinpath('beam_elements/elements_src/track_dipole_edge_linear.h'),
        _pkg_root.joinpath('beam_elements/elements_src/track_mult_fringe.h'),
        _pkg_root.joinpath('beam_elements/elements_src/track_magnet_edge.h'),
        _pkg_root.joinpath('beam_elements/elements_src/magnet_edge.h'),
    ]

    _repr_fields = [
        'model', 'is_exit', 'kn', 'ks', 'k_order', 'knl', 'ksl', 'kl_order',
        'length', 'half_gap', 'face_angle', 'face_angle_feed_down',
        'fringe_integral', 'delta_taper',
    ]

    _noexpr_fields = _NOEXPR_FIELDS

    def __init__(self, **kwargs):
        if '_xobject' in kwargs.keys() and kwargs['_xobject'] is not None:
            self.xoinitialize(**kwargs)
            return

        model = kwargs.pop('model', None)

        k_order = kwargs.pop('k_order', -1)
        kn, ks = kwargs.pop('kn', []), kwargs.pop('ks', [])
        k_multipolar_kwargs = _prepare_multipolar_params(
            k_order, skip_factorial=True, order_name='k_order', kn=kn, ks=ks)
        kwargs.update(k_multipolar_kwargs)

        kl_order = kwargs.pop('kl_order', -1)
        knl, ksl = kwargs.pop('knl', []), kwargs.pop('ksl', [])
        kl_multipolar_kwargs = _prepare_multipolar_params(
            kl_order, skip_factorial=True, order_name='kl_order', knl=knl, ksl=ksl)
        kwargs.update(kl_multipolar_kwargs)

        self.xoinitialize(**kwargs)

        if model is not None:
            self.model = model

    @property
    def model(self):
        return _INDEX_TO_EDGE_MODEL[self._model]

    @model.setter
    def model(self, value):
        try:
            self._model = _EDGE_MODEL_TO_INDEX[value]
        except KeyError:
            raise ValueError(f'Invalid edge model: {value}')

    def to_dict(self, copy_to_cpu=True):
        out = super().to_dict(copy_to_cpu=copy_to_cpu)

        if f'_model' in out:
            out.pop(f'_model')
        out['model'] = getattr(self, 'model')

        # See the comment in Multiple.to_dict about knl/ksl/order dumping
        for field in ['knl', 'ksl', 'kn', 'ks']:
            if field in out and np.allclose(out[field], 0, atol=1e-16):
                out.pop(field, None)

        if self.kl_order != -1 and 'knl' not in out and 'ksl' not in out:
            out['kl_order'] = self.order

        if self.k_order != -1 and 'kn' not in out and 'ks' not in out:
            out['k_order'] = self.order

        out['is_exit'] = bool(out['is_exit'])

        return out


def _prepare_multipolar_params(
        order=None,
        skip_factorial=False,
        order_name='order',
        **kwargs,
):
    """Prepare the multipolar parameters for an element with kicks.

    This function takes the multipolar coefficients and the order, and extends/
    computes new arrays of compatible order, padding them with zeros if needed.

    Parameters
    ----------
    order : int, optional
        The multipolar order. If not provided, will be inferred from the other
        parameters.
    order_name : str, optional
        The name of the field in ``kwargs`` that stores the order.
    skip_factorial : bool, optional
        Whether to calculate ``inv_factorial_order``. Skipped by default.
    kwargs : dict
        A dictionary with values that are either array-type fields that contain
        multipolar coefficients, or None.

    Returns
    -------
    dict
        A dictionary containing the order field named appropriately and the
        arrays given in ``kwargs``. The arrays will be extended with zeros (and
        None will spawn zeroed arrays) compatible with the given order. If
        ``order`` is not given, its value will be inferred from the given
        arrays. If ``skip_factorial`` is False, the returned dictionary will
        also contain ``inv_factorial_order``.
    """
    order = order or 0

    lengths = [len(kwarg) if kwarg is not None else 0 for kwarg in kwargs.values()]

    target_len = max((order + 1), *lengths)
    assert target_len >= 0

    new_kwargs = {}
    for kwarg_name, kwarg in kwargs.items():
        new_kwarg = np.zeros(target_len, dtype=np.float64)
        new_kwargs[kwarg_name] = new_kwarg
        if kwarg is None:
            continue
        if hasattr(kwarg, 'get'):
            kwarg = kwarg.get()
        new_kwarg[: len(kwarg)] = np.array(kwarg)

    order = target_len - 1

    new_kwargs[order_name] = order

    if not skip_factorial:
        new_kwargs['inv_factorial_order'] = 1.0 / factorial(order, exact=True)

    return new_kwargs


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xtrack as xt

import matplotlib.pyplot as plt

#################################
# Load a line and build tracker #
#################################

line = xt.Line.from_json(
    '../../test_data/lhc_no_bb/line_and_particle.json')
line.particle_ref = xt.Particles(
                    mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.build_tracker()

tw = line.twiss()
tw_htg = line.twiss(hide_thin_groups=True)

plt.close('all')
plt.plot(tw.s, tw.x, label='hide_thin_groups=False')
plt.plot(tw.s, tw_htg.x, label='hide_thin_groups=True')
plt.xlabel('s [m]')
plt.ylabel('x [m]')
plt.legend()
plt.show()

for nn in ('x y px py zeta delta ptau '
           'betx bety alfx alfy gamx gamy dx dy dpx dpy').split():
    assert np.isnan(tw_htg[nn]).sum() == 2281
    assert np.isnan(tw[nn]).sum() == 0

    # Check in presence of a srotation
    assert tw.name[11197] == 'mbxws.1r8_pretilt'
    assert tw.name[11198] == 'mbxws.1r8'
    assert tw.name[11199] == 'mbxws.1r8_posttilt'

    assert tw_htg[nn][11197] == tw[nn][11197]
    assert np.isnan(tw_htg[nn][11198])
    assert np.isnan(tw_htg[nn][11199])
    assert tw_htg[nn][11200] == tw[nn][11200]




import xtrack as xt

mad_source = '''
a = 3;
b = 3*a;  ! <<<< assiged by value
c := a^5; ! <<<< deferred expression

ll: sequence, l = 1.0;
mm: marker, at = 0.5;
endsequence;

'''

env_native = xt.load_madx_lattice(string=mad_source)

from cpymad.madx import Madx
mad = Madx()
mad.input(mad_source)
mad.beam()
mad.use(sequence='ll')
env_cpymad = xt.Line.from_madx_sequence(mad.sequence.ll,
                                        deferred_expressions=True).env

env_native.get_expr('c')
# (vars['a'] ** 5.0) <<<<<<< CORRECT
env_cpymad.get_expr('c')
# is vars['a^5'] <<<<<<<<<<< WRONG!

env_native.get_expr('b')
# (3.0 * vars['a']) <<<<<<< WRONG!, was assinged by value in madx
env_cpymad.get_expr('b')
# None <<<<<<<<<<< CORRECT


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import xobjects as xo

import xtrack as xt


class BeamInteraction:

    skip_in_loss_location_refinement = True

    def __init__(self, name=None, interaction_process=None,
                 length=0, isthick=None):

        self.name = name
        self.interaction_process = interaction_process

        self.length = length

        if isthick is None:
            isthick = True if length > 0 else False
        self.isthick = isthick

    def track(self, particles):

        assert isinstance(particles._buffer.context, xo.ContextCpu)
        assert particles._num_active_particles >= 0

        # Assumes active particles are contiguous
        products = self.interaction_process.interact(particles)

        # TODO: This should work also when no products are there
        #       Particles reorganization should still happen

        if products is None or products['x'].size == 0:
            particles.reorganize()
        else:
            new_particles = xt.Particles(_context=particles._buffer.context,
                    p0c = particles.p0c[0], # TODO: Should we check that 
                                            #       they are all the same?
                    mass0 = particles.mass0,
                    q0 = particles.q0,
                    s = products['s'],
                    x = products['x'],
                    px = products['px'],
                    y = products['y'],
                    py = products['py'],
                    zeta = products['zeta'],
                    delta = products['delta'],
                    mass_ratio = products['mass_ratio'],
                    charge_ratio = products['charge_ratio'],
                    at_element = products['at_element'],
                    at_turn = products['at_turn'],
                    parent_particle_id = products['parent_particle_id'])

            particles.add_particles(new_particles)

class ParticlesInjectionSample:

    def __init__(self, particles_to_inject, line, element_name, num_particles_to_inject):
        self.particles_to_inject = particles_to_inject.copy()
        self.line = line
        self.element_name = element_name
        self.num_particles_to_inject = num_particles_to_inject

    def track(self, particles):

        if not isinstance(particles._context, xo.ContextCpu):
            raise ValueError('This element only works with CPU context')

        self.particles_to_inject.at_turn += 1

        if not self.num_particles_to_inject:
            return

        assert self.element_name in self.line.element_names
        assert self.line.element_dict[self.element_name] is self
        at_element = self.line.element_names.index(self.element_name)
        s_element = self.line.tracker._tracker_data_base.element_s_locations[at_element]

        # Get random particles to inject
        idx_inject = np.random.default_rng().choice(len(self.particles_to_inject.x),
                                size=self.num_particles_to_inject, replace=False)

        mask_inject = np.zeros(len(self.particles_to_inject.x), dtype=bool)
        mask_inject[idx_inject] = True

        p_inj = self.particles_to_inject.filter(mask_inject)
        p_inj.s = s_element
        p_inj.at_element = at_element
        p_inj.update_p0c_and_energy_deviations(p0c=self.line.particle_ref.p0c[0])

        particles.add_particles(p_inj)


import numpy as np
import pytest

import xtrack as xt
import xobjects as xo
from xobjects.test_helpers import for_all_test_contexts

from xtrack.beam_elements.magnets import Magnet, MagnetEdge


def make_particles(context):
    return xt.Particles(
        kinetic_energy0=50e6,
        x=[1e-3, -1e-3],
        y=2e-3,
        zeta=1e-2,
        px=10e-3,
        py=20e-3,
        delta=1e-2,
        _context=context,
    )


@for_all_test_contexts
def test_magnet_expanded_drift(test_context):
    magnet = Magnet(
        length=2.0,
        k0=0.0,
        k1=0.0,
        h=0.0,
        integrator='teapot',
        model='drift-kick-drift-expanded',
        _context=test_context,
    )

    assert magnet.model == 'drift-kick-drift-expanded'
    assert magnet.integrator == 'teapot'

    drift = xt.Drift(length=2.0, _context=test_context)

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    magnet.track(p_test)
    drift.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_exact_drift(test_context):
    magnet = Magnet(
        length=2.0,
        k0=0.0,
        k1=0.0,
        h=0.0,
        integrator='teapot',
        model='drift-kick-drift-exact',
        _context=test_context,
    )

    assert magnet.model == 'drift-kick-drift-exact'
    assert magnet.integrator == 'teapot'

    exact_drift = xt.Solenoid(length=2.0, _context=test_context)  # Solenoid is exact drift when off

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    magnet.track(p_test)
    exact_drift.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[magnet])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_sextupole(test_context):
    magnet = Magnet(
        length=2.0,
        k2=3,
        k2s=5,
        integrator='teapot',
        model='drift-kick-drift-expanded',
        _context=test_context,
    )

    sextupole = xt.Sextupole(length=2.0, k2=3., k2s=5., _context=test_context)
    assert magnet.integrator == 'teapot'
    assert magnet.model == 'drift-kick-drift-expanded'

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    magnet.track(p_test)
    sextupole.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[magnet])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_sextupole_with_kicks_that_do_nothing(test_context):
    magnet = Magnet(
        length=2.0,
        k2=3,
        k2s=5,
        num_multipole_kicks=5,
        integrator='uniform',
        model='drift-kick-drift-expanded',
        _context=test_context,
    )

    sextupole = xt.Sextupole(
        length=2.0,
        k2=3.,
        k2s=5.,
        num_multipole_kicks=5,
        _context=test_context,
    )

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    magnet.track(p_test)
    sextupole.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[magnet])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_sextupole_with_kicks(test_context):
    magnet = Magnet(
        length=2.0,
        k2s=5,
        knl=[0., 0., 3. * 2., 0., 0., 0.],
        num_multipole_kicks=5,
        integrator='teapot',
        model='drift-kick-drift-expanded',
        _context=test_context,
    )

    sextupole = xt.Sextupole(
        length=2.0,
        k2=3.,
        k2s=5.,
        num_multipole_kicks=5,
        integrator='teapot',
        _context=test_context,
    )
    assert sextupole.integrator == 'teapot'

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    magnet.track(p_test)
    sextupole.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[magnet])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_sextupole_with_skew_kick(test_context):
    magnet = Magnet(
        length=2.0,
        integrator='uniform',
        model='drift-kick-drift-expanded',
        k2=0,
        k2s=5,
        knl=[0, 0, 6, 0, 0, 0],
        ksl=[0, 0, -2, 0, 0, 0],
        num_multipole_kicks=5,
        _context=test_context,
    )

    sextupole = xt.Sextupole(
        length=2.0,
        k2=3,
        k2s=5,
        num_multipole_kicks=5,
        _context=test_context,
    )
    sextupole.ksl[2] = -2.

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    magnet.track(p_test)
    sextupole.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[magnet])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
@pytest.mark.parametrize('integrator', ['adaptive', 'teapot'])
def test_magnet_quadrupole(test_context, integrator):
    magnet = Magnet(
        length=2.0,
        k1=3,
        num_multipole_kicks=1,
        integrator='teapot',
        model='mat-kick-mat',
        _context=test_context,
    )

    quadrupole = xt.Quadrupole(
        length=2,
        k1=3,
        num_multipole_kicks=1,
        integrator=integrator,
        _context=test_context,
    )
    assert quadrupole.integrator == integrator

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    magnet.track(p_test)
    quadrupole.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[magnet])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_curved_quad(test_context):
    magnet = Magnet(
        length=2.0,
        h=0.05,
        k1=-0.3,
        num_multipole_kicks=15,
        integrator='yoshida4',
        model='rot-kick-rot',
        _context=test_context,
    )

    bend = xt.Bend(
        length=2.0,
        k1=-0.3,
        h=0.05,
        num_multipole_kicks=15,
        _context=test_context,
    )

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    magnet.track(p_test)
    bend.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[magnet])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_bend_auto_no_kicks(test_context):
    magnet = Magnet(
        length=2.0,
        h=0.05,
        k1=0,
        model='bend-kick-bend',
        integrator='yoshida4',
        num_multipole_kicks=0,
        _context=test_context,
    )

    eref = xt.Bend(
        length=2.0,
        h=0.05,
        num_multipole_kicks=0,
        _context=test_context,
    )

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    magnet.track(p_test)
    eref.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[magnet])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_bend_auto_quad_kick(test_context):
    magnet = Magnet(
        length=2.0,
        model='bend-kick-bend',
        integrator='yoshida4',
        num_multipole_kicks=1,
        h=0.05,
        k1=0.3,
        _context=test_context,
    )

    bend = xt.Bend(
        length=2.0,
        h=0.05,
        k1=0.3,
        model='bend-kick-bend',
        num_multipole_kicks=1,
        edge_entry_active=False,
        edge_exit_active=False,
        _context=test_context,
    )

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    magnet.track(p_test)
    bend.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[magnet])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
@pytest.mark.parametrize('model', ['bend-kick-bend', 'rot-kick-rot', 'expanded'])
def test_magnet_bend_dip_quad_kick(model, test_context):
    magnet = Magnet(
        length=2.0,
        k0=0.2,
        k1=0.3,
        h=0.1,
        integrator='yoshida4',
        num_multipole_kicks=10,
        _context=test_context,
    )

    bend = xt.Bend(
        length=2.0,
        h=0.1,
        k1=0.3,
        k0=0.2,
        num_multipole_kicks=10,
        _context=test_context,
    )
    bend.edge_entry_active = False
    bend.edge_exit_active = False

    magnet.model = model
    bend.model = model

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    magnet.track(p_test)
    bend.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[magnet])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, rtol=0, atol=1e-7)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
@pytest.mark.parametrize('model', ['bend-kick-bend', 'rot-kick-rot', 'expanded'])
def test_magnet_bend_dip_quad_kick_with_multipoles(model, test_context):
    magnet = Magnet(
        length=2.0,
        h=0.1,
        k0=0.2,
        k1=0.3,
        k2=0.1,
        k3=0.15,
        k0s=0.02,
        k1s=0.03,
        k2s=0.01,
        k3s=0.02,
        knl=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
        ksl=[0.6, 0.5, 0.4, 0.15, 0.2, 0.1],
        num_multipole_kicks=10,
        integrator='yoshida4',
        _context=test_context,
    )

    bend = xt.Bend(
        length=2.0,
        h=0.1,
        k1=0.3,
        k0=0.2,
        num_multipole_kicks=10,
        _context=test_context,
    )
    bend.edge_entry_active = False
    bend.edge_exit_active = False
    bend.knl = test_context.nparray_to_context_array(
        np.array([0.1, 0.2, 0.3 + 0.1 * 2, 0.4 + 0.15 * 2, 0.5, 0.6])
    )
    bend.ksl = test_context.nparray_to_context_array(
        np.array([0.6 + 0.02 * 2, 0.5 + 0.03 * 2, 0.4 + 0.01 * 2, 0.15 + 0.02 * 2, 0.2, 0.1])
    )

    magnet.model = model
    bend.model = model

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    magnet.track(p_test)
    bend.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-14, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[magnet])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, rtol=0, atol=1e-7)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_check_uniform_integrator(test_context):
    mm1 = Magnet(h=0.1, k1=0.3, k0=0.2, length=2.0, _context=test_context)
    mm2 = mm1.copy()

    mm1.edge_entry_active = False
    mm1.edge_exit_active = False
    mm2.edge_entry_active = False
    mm2.edge_exit_active = False

    mm1.integrator = 'uniform'
    mm2.integrator = 'teapot'
    mm1.num_multipole_kicks = 1
    mm2.num_multipole_kicks = 1

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    mm1.track(p_test)
    mm2.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[mm1])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

    # more kicks (needs loser thresholds)
    mm1.num_multipole_kicks = 10
    mm2.num_multipole_kicks = 10

    p_test = p0.copy()
    p_ref = p0.copy()

    mm1.track(p_test)
    mm2.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref_cpu.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=0, rtol=5e-3)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=0, rtol=5e-3)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=0, rtol=1e-2)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=0, rtol=5e-3)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=0, rtol=5e-3)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=0, rtol=5e-3)

    # Test backtracking
    line = xt.Line(elements=[mm1])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts(excluding='ContextPyopencl')
def test_edge_suppressed_edge(test_context):
    e_test = MagnetEdge(model='suppressed', kn=[0], ks=[0], _context=test_context)
    e_ref = xt.DipoleEdge(model='suppressed', k=0, _context=test_context)

    p0 = xt.Particles(
        kinetic_energy0=50e6,
        x=1e-2, y=2e-2, zeta=1e-2, px=10e-2, py=20e-2, delta=1e-2,
        _context=test_context,
    )

    # Expanded drift
    p_test = p0.copy()
    p_ref = p0.copy()

    e_test.track(p_test)
    e_ref.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts(excluding='ContextPyopencl')
def test_edge_linear_edge_does_nothing(test_context):
    e_test = MagnetEdge(model='linear', kn=[0], ks=[0], _context=test_context)
    e_ref = xt.DipoleEdge(model='linear', k=0, _context=test_context)

    p0 = xt.Particles(
        kinetic_energy0=50e6,
        x=1e-2, y=2e-2, zeta=1e-2, px=10e-2, py=20e-2, delta=1e-2,
        _context=test_context,
    )

    # Expanded drift
    p_test = p0.copy()
    p_ref = p0.copy()

    e_test.track(p_test)
    e_ref.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts(excluding='ContextPyopencl')
def test_edge_full_edge_does_nothing(test_context):
    e_test = MagnetEdge(model='full', kn=[0], ks=[0], _context=test_context)
    e_ref = xt.DipoleEdge(model='full', k=0, _context=test_context)

    p0 = xt.Particles(
        kinetic_energy0=50e6,
        x=1e-2, y=2e-2, zeta=1e-2, px=10e-2, py=20e-2, delta=1e-2,
        _context=test_context,
    )

    # Expanded drift
    p_test = p0.copy()
    p_ref = p0.copy()

    e_test.track(p_test)
    e_ref.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts(excluding='ContextPyopencl')
def test_edge_only_linear_edge(test_context):
    e_test = MagnetEdge(
        model='linear', kn=[3], face_angle=0.1, face_angle_feed_down=0.2,
        fringe_integral=0.3,
        half_gap=0.4, _context=test_context
    )
    e_ref = xt.DipoleEdge(
        model='linear', k=3, e1=0.1, e1_fd=0.2, fint=0.3, hgap=0.4,
        _context=test_context
    )

    p0 = xt.Particles(
        kinetic_energy0=50e6,
        x=1e-2, y=2e-2, zeta=1e-2, px=10e-2, py=20e-2, delta=1e-2,
        _context=test_context,
    )

    # Expanded drift
    p_test = p0.copy()
    p_ref = p0.copy()

    e_test.track(p_test)
    e_ref.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())
    
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts(excluding='ContextPyopencl')
def test_edge_full_edge_with_dipole_component(test_context):
    e_test = MagnetEdge(
        model='full', kn=[3], face_angle=0.1, face_angle_feed_down=0.2,
        fringe_integral=0.3,
        half_gap=0.4, _context=test_context
    )
    e_ref = xt.DipoleEdge(
        model='full', k=3, e1=0.1, e1_fd=0.2, fint=0.3, hgap=0.4,
        _context=test_context
    )

    p0 = xt.Particles(
        kinetic_energy0=50e6,
        x=1e-2, y=2e-2, zeta=1e-2, px=10e-2, py=20e-2, delta=1e-2,
        _context=test_context,
    )

    # Expanded drift
    p_test = p0.copy()
    p_ref = p0.copy()

    e_test.track(p_test)
    e_ref.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts(excluding='ContextPyopencl')
def test_edge_multipole_fringe_without_dipole_component(test_context):
    e_test = MagnetEdge(
        model='full', kn=[0, 2, 3], k_order=2, _context=test_context
    )
    e_ref = xt.MultipoleEdge(kn=[0, 2, 3], order=2, _context=test_context)

    p0 = xt.Particles(
        kinetic_energy0=50e6,
        x=1e-2, y=2e-2, zeta=1e-2, px=10e-2, py=20e-2, delta=1e-2,
        _context=test_context,
    )

    # Expanded drift
    p_test = p0.copy()
    p_ref = p0.copy()

    e_test.track(p_test)
    e_ref.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())
    
    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts(excluding='ContextPyopencl')
def test_edge_full_model_with_dipole_component_no_angle(test_context):
    e_test = MagnetEdge(
        model='full', kn=[3, 4, 5], fringe_integral=0.3, half_gap=0.4, k_order=2,
        _context=test_context
    )
    e_ref = [
        xt.DipoleEdge(model='full', k=3, fint=0.3, hgap=0.4),
        xt.MultipoleEdge(kn=[0, 4, 5], order=2),
    ]

    p0 = xt.Particles(
        kinetic_energy0=50e6,
        x=1e-2, y=2e-2, zeta=1e-2, px=10e-2, py=20e-2, delta=1e-2,
        _context=test_context,
    )

    # Expanded drift
    p_test = p0.copy()
    p_ref = p0.copy()

    e_test.track(p_test)

    mini_line = xt.Line(elements=e_ref)
    mini_line.build_tracker(_context=test_context)
    mini_line.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts(excluding='ContextPyopencl')
def test_edge_full_model_with_dipole_component_and_angle(test_context):
    e_test = MagnetEdge(
        model='full', kn=[3, 4, 5], face_angle=0.2, face_angle_feed_down=0.0,
        fringe_integral=0.3,
        half_gap=0.4, k_order=2, _context=test_context
    )
    e_ref = [
        xt.YRotation(angle=np.rad2deg(-0.2)),
        # The rotation is also the other way than in the underlying map :'(
        xt.DipoleEdge(model='full', k=3, fint=0.3, hgap=0.4),
        xt.MultipoleEdge(kn=[0, 4, 5], order=2),
        xt.Wedge(angle=-0.2, k=3),
    ]

    p0 = xt.Particles(
        kinetic_energy0=50e6,
        x=1e-2, y=2e-2, zeta=1e-2, px=10e-2, py=20e-2, delta=1e-2,
        _context=test_context,
    )

    # Expanded drift
    p_test = p0.copy()
    p_ref = p0.copy()

    e_test.track(p_test)

    mini_line = xt.Line(elements=e_ref)
    mini_line.build_tracker(_context=test_context)
    mini_line.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts(excluding='ContextPyopencl')
def test_edge_full_model_with_dipole_component_and_angle_exit(test_context):
    e_test = MagnetEdge(
        model='full', kn=[3, 4, 5], is_exit=True, face_angle=0.2,
        face_angle_feed_down=0.0,
        fringe_integral=0.3, half_gap=0.4, k_order=2, _context=test_context
    )
    e_ref = [
        xt.Wedge(angle=-0.2, k=3),
        xt.MultipoleEdge(kn=[0, 4, 5], is_exit=True, order=2),
        xt.DipoleEdge(model='full', k=-3, fint=0.3, hgap=0.4),
        xt.YRotation(angle=np.rad2deg(-0.2)),
    ]

    p0 = xt.Particles(
        kinetic_energy0=50e6,
        x=1e-2, y=2e-2, zeta=1e-2, px=10e-2, py=20e-2, delta=1e-2,
        _context=test_context,
    )

    # Expanded drift
    p_test = p0.copy()
    p_ref = p0.copy()

    e_test.track(p_test)

    mini_line = xt.Line(elements=e_ref)
    mini_line.build_tracker(_context=test_context)
    mini_line.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts(excluding='ContextPyopencl')
def test_edge_linear_edge_exit(test_context):
    e_test = MagnetEdge(
        model='linear', is_exit=True, kn=[3], face_angle=0.1,
        face_angle_feed_down=0.2,
        fringe_integral=0.3, half_gap=0.4, _context=test_context
    )
    e_ref = xt.DipoleEdge(
        model='linear', k=3, e1=0.1, e1_fd=0.2, fint=0.3, hgap=0.4,
        _context=test_context,
    )

    p0 = xt.Particles(
        kinetic_energy0=50e6,
        x=1e-2, y=2e-2, zeta=1e-2, px=10e-2, py=20e-2, delta=1e-2,
        _context=test_context,
    )

    # Expanded drift
    p_test = p0.copy()
    p_ref = p0.copy()

    e_test.track(p_test)
    e_ref.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_and_edge_only_linear(test_context):
    bb = xt.Bend(
        h=0.1, k0=0.11, length=0,
        edge_entry_angle=0.02, edge_exit_angle=0.03,
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,
        edge_entry_fint=0.1, edge_exit_fint=0.2,
        _context=test_context,
    )

    bb.edge_entry_active = 1
    bb.edge_exit_active = 0
    bb.model = 'rot-kick-rot'
    bb.num_multipole_kicks = 10

    mm = Magnet(
        h=0.1, k0=0.11, length=0,
        edge_entry_angle=0.02, edge_exit_angle=0.03,
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,
        edge_entry_fint=0.1, edge_exit_fint=0.2,
        _context=test_context,
    )
    mm.edge_entry_active = 1
    mm.edge_exit_active = 0
    mm.model = 'rot-kick-rot'
    mm.num_multipole_kicks = 10

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    mm.track(p_test)
    bb.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[mm])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_and_edge_exit_alone(test_context):
    bb = xt.Bend(
        h=0.1, k0=0.11, length=0,
        edge_entry_angle=0.02, edge_exit_angle=0.03,
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,
        edge_entry_fint=0.1, edge_exit_fint=0.2,
        _context=test_context,
    )

    bb.edge_entry_active = 0
    bb.edge_exit_active = 1
    bb.model = 'rot-kick-rot'
    bb.num_multipole_kicks = 10

    mm = Magnet(
        h=0.1, k0=0.11, length=0,
        edge_entry_angle=0.02, edge_exit_angle=0.03,
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,
        edge_entry_fint=0.1, edge_exit_fint=0.2,
        _context=test_context,
    )

    mm.edge_entry_active = 0
    mm.edge_exit_active = 1
    mm.model = 'rot-kick-rot'
    mm.num_multipole_kicks = 10

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    mm.track(p_test)
    bb.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[mm])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_and_edge_full_bend_linear_edges(test_context):
    bb = xt.Bend(
        h=0.1, k0=0.11, length=10,
        edge_entry_angle=0.02, edge_exit_angle=0.03,
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,
        edge_entry_fint=0.1, edge_exit_fint=0.2,
        _context=test_context,
    )

    bb.edge_entry_active = 1
    bb.edge_exit_active = 1
    bb.model = 'rot-kick-rot'
    bb.num_multipole_kicks = 10
    bb.edge_entry_model = 'linear'
    bb.edge_exit_model = 'linear'

    mm = Magnet(
        h=0.1, k0=0.11, length=10,
        edge_entry_angle=0.02, edge_exit_angle=0.03,
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,
        edge_entry_fint=0.1, edge_exit_fint=0.2,
        _context=test_context,
    )

    mm.edge_entry_active = 1
    mm.edge_exit_active = 1
    mm.model = 'rot-kick-rot'
    mm.num_multipole_kicks = 10
    mm.edge_entry_model = 'linear'
    mm.edge_exit_model = 'linear'

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    mm.track(p_test)
    bb.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-13, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[mm])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_and_edge_nonlinear_entry_alone(test_context):
    bb = xt.Bend(
        h=0.1, k0=0.11, length=0,
        edge_entry_angle=0.02, edge_exit_angle=0.03,
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,
        edge_entry_fint=0.1, edge_exit_fint=0.2,
        _context=test_context,
    )

    bb.edge_entry_active = 1
    bb.edge_exit_active = 0
    bb.model = 'rot-kick-rot'
    bb.num_multipole_kicks = 10
    bb.edge_entry_model = 'full'

    mm = Magnet(
        h=0.1, k0=0.11, length=0,
        edge_entry_angle=0.02, edge_exit_angle=0.03,
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,
        edge_entry_fint=0.1, edge_exit_fint=0.2,
        _context=test_context,
    )

    mm.edge_entry_active = 1
    mm.edge_exit_active = 0
    mm.model = 'rot-kick-rot'
    mm.num_multipole_kicks = 10
    mm.edge_entry_model = 'full'

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    mm.track(p_test)
    bb.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_and_edge_nonlinear_exit_alone(test_context):
    bb = xt.Bend(
        h=0.1, k0=0.11, length=0,
        edge_entry_angle=0.02, edge_exit_angle=0.03,
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,
        edge_entry_fint=0.1, edge_exit_fint=0.2,
        _context=test_context,
    )

    bb.edge_entry_active = 0
    bb.edge_exit_active = 1
    bb.model = 'rot-kick-rot'
    bb.num_multipole_kicks = 10
    bb.edge_exit_model = 'full'

    mm = Magnet(
        h=0.1, k0=0.11, length=0,
        edge_entry_angle=0.02, edge_exit_angle=0.03,
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,
        edge_entry_fint=0.1, edge_exit_fint=0.2,
        _context=test_context,
    )

    mm.edge_entry_active = 0
    mm.edge_exit_active = 1
    mm.model = 'rot-kick-rot'
    mm.num_multipole_kicks = 10
    mm.edge_exit_model = 'full'

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    mm.track(p_test)
    bb.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_and_edge_nonlinear_both_edges(test_context):
    bb = xt.Bend(
        h=0.1, k0=0.11, length=10,
        edge_entry_angle=0.02, edge_exit_angle=0.03,
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,
        edge_entry_fint=0.1, edge_exit_fint=0.2,
        _context=test_context,
    )

    bb.edge_entry_active = 1
    bb.edge_exit_active = 1
    bb.model = 'rot-kick-rot'
    bb.num_multipole_kicks = 10
    bb.edge_entry_model = 'full'
    bb.edge_exit_model = 'full'

    mm = Magnet(
        h=0.1, k0=0.11, length=10,
        edge_entry_angle=0.02, edge_exit_angle=0.03,
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,
        edge_entry_fint=0.1, edge_exit_fint=0.2,
        _context=test_context,
    )

    mm.edge_entry_active = 1
    mm.edge_exit_active = 1
    mm.model = 'rot-kick-rot'
    mm.num_multipole_kicks = 10
    mm.edge_entry_model = 'full'
    mm.edge_exit_model = 'full'

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    mm.track(p_test)
    bb.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=3e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_and_edge_quadrupole_nonlinear_fringes(test_context):
    qq = xt.Quadrupole(k1=0.11, length=3, _context=test_context)
    qq.edge_entry_active = 1
    qq.edge_exit_active = 1

    mm = Magnet(
        k0=0, k1=0.11, length=3,
        edge_entry_model='full', edge_exit_model='full',
        edge_entry_fint=0.1, edge_exit_fint=0.2,  # should be ignored
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,  # should be ignored
        _context=test_context,
    )
    mm.edge_entry_active = 1
    mm.edge_exit_active = 1
    mm.model = 'mat-kick-mat'

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    mm.track(p_test)
    qq.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_and_edge_sextupole_nonlinear_fringes(test_context):
    ss = xt.Sextupole(k2=0.11 + 0.03 / 3., k2s=0.12 - 0.04 / 3, length=3, _context=test_context)
    ss.edge_entry_active = 1
    ss.edge_exit_active = 1

    mm = Magnet(
        k0=0, k2=0.11, k2s=0.12, length=3,
        knl=[0, 0, 0.03], ksl=[0, 0, -0.04],
        edge_entry_model='full', edge_exit_model='full',
        edge_entry_fint=0.1, edge_exit_fint=0.2,  # should be ignored
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,  # should be ignored
        _context=test_context,
    )
    mm.edge_entry_active = 1
    mm.edge_exit_active = 1
    mm.model = 'drift-kick-drift-expanded'
    mm.num_multipole_kicks = 1
    mm.integrator = 'uniform'

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    mm.track(p_test)
    ss.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)


@for_all_test_contexts
def test_magnet_and_edge_octupole_nonlinear_fringes(test_context):
    oo = xt.Octupole(k3=0.3 + 0.02 / 3., k3s=0.4 - 0.07 / 3, length=3, _context=test_context)
    oo.edge_entry_active = 1
    oo.edge_exit_active = 1
    oo.edge_entry_active = 1
    oo.edge_exit_active = 1

    mm = Magnet(
        k0=0, k3=0.3, k3s=0.4, length=3,
        knl=[0, 0, 0, 0.02], ksl=[0, 0, 0, -0.07],
        edge_entry_model='full', edge_exit_model='full',
        edge_entry_fint=0.1, edge_exit_fint=0.2,  # should be ignored
        edge_entry_hgap=0.04, edge_exit_hgap=0.05,  # should be ignored
        _context=test_context,
    )
    mm.edge_entry_active = 1
    mm.edge_exit_active = 1
    mm.model = 'drift-kick-drift-expanded'
    mm.num_multipole_kicks = 1
    mm.integrator = 'uniform'

    p0 = make_particles(test_context)
    p_test = p0.copy()
    p_ref = p0.copy()

    mm.track(p_test)
    oo.track(p_ref)

    p_test_cpu = p_test.copy(_context=xo.ContextCpu())
    p_ref_cpu = p_ref.copy(_context=xo.ContextCpu())

    xo.assert_allclose(p_test_cpu.x, p_ref_cpu.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.y, p_ref_cpu.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.zeta, p_ref_cpu.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.px, p_ref_cpu.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.py, p_ref_cpu.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test_cpu.delta, p_ref_cpu.delta, atol=1e-15, rtol=0)

    line = xt.Line(elements=[mm])
    line.build_tracker(compile=False, _context=test_context)
    line.track(p_test, backtrack=True)
    p_test.move(_context=xo.ContextCpu())
    assert np.all(p_test.state == -32)

def test_bend_convergence_on_axis():

    bb = xt.Bend(k0=0.001, h=0.001, length=2)
    bb.integrator = 'yoshida4'
    bb.num_multipole_kicks = 20

    p0 = xt.Particles(x=0.0, y=0.0, delta=[0, 1e-3])

    bb.model = 'bend-kick-bend'
    assert bb._xobject.model == 2
    p_bkb = p0.copy()
    bb.track(p_bkb)

    bb.model = 'rot-kick-rot'
    assert bb._xobject.model == 3
    p_rkr = p0.copy()
    bb.track(p_rkr)

    bb.model = 'mat-kick-mat'
    assert bb._xobject.model == 4
    p_mkm = p0.copy()
    bb.track(p_mkm)

    bb.model = 'drift-kick-drift-exact'
    assert bb._xobject.model == 5
    p_dkd1 = p0.copy()
    bb.track(p_dkd1)

    bb.model = 'drift-kick-drift-expanded'
    assert bb._xobject.model == 6
    p_dkd2 = p0.copy()
    bb.track(p_dkd2)

    xo.assert_allclose(p_bkb.x, p_rkr.x, rtol=0, atol=1e-12)
    xo.assert_allclose(p_bkb.x, p_mkm.x, rtol=0, atol=1e-12)
    xo.assert_allclose(p_bkb.x, p_dkd1.x, rtol=0, atol=1e-12)
    xo.assert_allclose(p_bkb.x, p_dkd2.x, rtol=0, atol=1e-12)

    xo.assert_allclose(p_bkb.px, p_rkr.px, rtol=0, atol=1e-14)
    xo.assert_allclose(p_bkb.px, p_mkm.px, rtol=0, atol=1e-14)
    xo.assert_allclose(p_bkb.px, p_dkd1.px, rtol=0, atol=1e-14)
    xo.assert_allclose(p_bkb.px, p_dkd2.px, rtol=0, atol=1e-14)

    xo.assert_allclose(p_bkb.y, p_rkr.y, rtol=0, atol=1e-12)
    xo.assert_allclose(p_bkb.y, p_mkm.y, rtol=0, atol=1e-12)
    xo.assert_allclose(p_bkb.y, p_dkd1.y, rtol=0, atol=1e-12)
    xo.assert_allclose(p_bkb.y, p_dkd2.y, rtol=0, atol=1e-12)

    xo.assert_allclose(p_bkb.py, p_rkr.py, rtol=0, atol=1e-14)
    xo.assert_allclose(p_bkb.py, p_mkm.py, rtol=0, atol=1e-14)
    xo.assert_allclose(p_bkb.py, p_dkd1.py, rtol=0, atol=1e-14)
    xo.assert_allclose(p_bkb.py, p_dkd2.py, rtol=0, atol=1e-14)

    xo.assert_allclose(p_bkb.zeta, p_rkr.zeta, rtol=0, atol=1e-14)
    xo.assert_allclose(p_bkb.zeta, p_mkm.zeta, rtol=0, atol=1e-14)
    xo.assert_allclose(p_bkb.zeta, p_dkd1.zeta, rtol=0, atol=1e-14)
    xo.assert_allclose(p_bkb.zeta, p_dkd2.zeta, rtol=0, atol=1e-14)

    xo.assert_allclose(p_bkb.delta, p_rkr.delta, rtol=0, atol=1e-14)
    xo.assert_allclose(p_bkb.delta, p_mkm.delta, rtol=0, atol=1e-14)
    xo.assert_allclose(p_bkb.delta, p_dkd1.delta, rtol=0, atol=1e-14)
    xo.assert_allclose(p_bkb.delta, p_dkd2.delta, rtol=0, atol=1e-14)

def test_convergence_mat_kick_mat():

    magnet = xt.Magnet(k0=0.02, h=0.01, k1=0.01, length=2.,
                    k2=0.005, k3=0.03,
                    k1s=0.01, k2s=0.005, k3s=0.05,
                    knl=[0.003, 0.001, 0.01, 0.02, 4., 6e2, 7e6],
                    ksl=[-0.005, 0.002, -0.02, 0.03, -2, 700., 4e6])

    p0 = xt.Particles(x=1e-2, y=2e-2, py=1e-3, delta=3e-2)

    m_ref = magnet.copy()
    m_ref.model = 'mat-kick-mat'
    m_ref.num_multipole_kicks = 1000
    p_ref = p0.copy()
    m_ref.track(p_ref)

    m_uniform = magnet.copy()
    m_uniform.model = 'drift-kick-drift-expanded'
    m_uniform.integrator='uniform'
    m_uniform.num_multipole_kicks = 50000

    m_teapot = magnet.copy()
    m_teapot.model = 'drift-kick-drift-expanded'
    m_teapot.integrator='teapot'
    m_teapot.num_multipole_kicks = 50000

    m_yoshida = magnet.copy()
    m_yoshida.model = 'drift-kick-drift-expanded'
    m_yoshida.integrator='yoshida4'
    m_yoshida.num_multipole_kicks = 500

    p_ref = p0.copy()
    p_uniform = p0.copy()
    p_teapot = p0.copy()
    p_yoshida = p0.copy()

    m_ref.track(p_ref)
    m_uniform.track(p_uniform)
    m_teapot.track(p_teapot)
    m_yoshida.track(p_yoshida)

    xo.assert_allclose(p_ref.x, p_uniform.x, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.px, p_uniform.px, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.y, p_uniform.y, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.py, p_uniform.py, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.zeta, p_uniform.zeta, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.delta, p_uniform.delta, rtol=0, atol=1e-13)

    xo.assert_allclose(p_ref.x, p_teapot.x, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.px, p_teapot.px, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.y, p_teapot.y, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.py, p_teapot.py, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.zeta, p_teapot.zeta, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.delta, p_teapot.delta, rtol=0, atol=1e-13)

    xo.assert_allclose(p_ref.x, p_yoshida.x, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.px, p_yoshida.px, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.y, p_yoshida.y, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.py, p_yoshida.py, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.zeta, p_yoshida.zeta, rtol=0, atol=1e-13)
    xo.assert_allclose(p_ref.delta, p_yoshida.delta, rtol=0, atol=1e-13)

def test_convergence_rot_kick_rot():

    magnet = xt.Magnet(k0=0.02, h=0.01, k1=0.01, length=2.,
                    k2=0.005, k3=0.03,
                    k1s=0.01, k2s=0.005, k3s=0.05,
                    knl=[0.003, 0.001, 0.01, 0.02, 4., 6e2, 7e6],
                    ksl=[-0.005, 0.002, -0.02, 0.03, -2, 700., 4e6])
    magnet.integrator = 'yoshida4'
    magnet.num_multipole_kicks = 50

    p0 = xt.Particles(x=1e-2, y=2e-2, py=1e-3, delta=3e-2)

    model_to_test = 'rot-kick-rot'

    m_ref = magnet.copy()
    m_ref.model = 'bend-kick-bend'
    p_ref = p0.copy()
    m_ref.track(p_ref)

    m_uniform = magnet.copy()
    m_uniform.model = model_to_test
    m_uniform.integrator='uniform'
    m_uniform.num_multipole_kicks = 50000

    m_teapot = magnet.copy()
    m_teapot.model = model_to_test
    m_teapot.integrator='teapot'
    m_teapot.num_multipole_kicks = 50000

    m_yoshida = magnet.copy()
    m_yoshida.model = model_to_test
    m_yoshida.integrator='yoshida4'
    m_yoshida.num_multipole_kicks = 100

    p_ref = p0.copy()
    p_uniform = p0.copy()
    p_teapot = p0.copy()
    p_yoshida = p0.copy()

    m_ref.track(p_ref)
    m_uniform.track(p_uniform)
    m_teapot.track(p_teapot)
    m_yoshida.track(p_yoshida)

    xo.assert_allclose(p_ref.x, p_uniform.x, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.px, p_uniform.px, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.y, p_uniform.y, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.py, p_uniform.py, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.zeta, p_uniform.zeta, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.delta, p_uniform.delta, rtol=0, atol=5e-13)

    xo.assert_allclose(p_ref.x, p_teapot.x, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.px, p_teapot.px, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.y, p_teapot.y, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.py, p_teapot.py, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.zeta, p_teapot.zeta, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.delta, p_teapot.delta, rtol=0, atol=5e-13)

    xo.assert_allclose(p_ref.x, p_yoshida.x, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.px, p_yoshida.px, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.y, p_yoshida.y, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.py, p_yoshida.py, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.zeta, p_yoshida.zeta, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.delta, p_yoshida.delta, rtol=0, atol=5e-13)

def test_convergence_drift_kick_drift_exact():

    magnet = xt.Magnet(k0=0.02, h=0., k1=0.01, length=2.,
                    k2=0.005, k3=0.03,
                    k1s=0.01, k2s=0.005, k3s=0.05,
                    knl=[0.003, 0.001, 0.01, 0.02, 4., 6e2, 7e6],
                    ksl=[-0.005, 0.002, -0.02, 0.03, -2, 700., 4e6])
    magnet.integrator = 'yoshida4'
    magnet.num_multipole_kicks = 100

    p0 = xt.Particles(x=1e-2, y=2e-2, py=1e-3, delta=3e-2)

    model_to_test = 'drift-kick-drift-exact'

    m_ref = magnet.copy()
    m_ref.model = 'bend-kick-bend'

    m_uniform = magnet.copy()
    m_uniform.model = model_to_test
    m_uniform.integrator='uniform'
    m_uniform.num_multipole_kicks = 50000

    m_teapot = magnet.copy()
    m_teapot.model = model_to_test
    m_teapot.integrator='teapot'
    m_teapot.num_multipole_kicks = 50000

    m_yoshida = magnet.copy()
    m_yoshida.model = model_to_test
    m_yoshida.integrator='yoshida4'
    m_yoshida.num_multipole_kicks = 100


    p_ref = p0.copy()
    p_uniform = p0.copy()
    p_teapot = p0.copy()
    p_yoshida = p0.copy()

    m_ref.track(p_ref)
    m_uniform.track(p_uniform)
    m_teapot.track(p_teapot)
    m_yoshida.track(p_yoshida)

    xo.assert_allclose(p_ref.x, p_uniform.x, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.px, p_uniform.px, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.y, p_uniform.y, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.py, p_uniform.py, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.zeta, p_uniform.zeta, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.delta, p_uniform.delta, rtol=0, atol=5e-13)

    xo.assert_allclose(p_ref.x, p_teapot.x, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.px, p_teapot.px, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.y, p_teapot.y, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.py, p_teapot.py, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.zeta, p_teapot.zeta, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.delta, p_teapot.delta, rtol=0, atol=5e-13)

    xo.assert_allclose(p_ref.x, p_yoshida.x, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.px, p_yoshida.px, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.y, p_yoshida.y, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.py, p_yoshida.py, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.zeta, p_yoshida.zeta, rtol=0, atol=5e-13)
    xo.assert_allclose(p_ref.delta, p_yoshida.delta, rtol=0, atol=5e-13)

def test_bend_expanded_exact_small_px():

    magnet = xt.Magnet(k0=0.002, h=0.002, k1=0.02, length=2)

    m_exact = magnet.copy()
    m_exact.model = 'bend-kick-bend'
    m_exact.integrator='yoshida4'
    m_exact.num_multipole_kicks = 1000

    m_expanded = magnet.copy()
    m_expanded.model = 'mat-kick-mat'
    m_expanded.integrator='yoshida4'
    m_expanded.num_multipole_kicks = 1000

    p0 = xt.Particles(x=1e-3, y=2e-3, px=5e-6)

    p_exact = p0.copy()
    m_exact.track(p_exact)

    p_expanded = p0.copy()
    m_expanded.track(p_expanded)

    xo.assert_allclose(p_exact.x, p_expanded.x, rtol=0, atol=1e-10)
    xo.assert_allclose(p_exact.px, p_expanded.px, rtol=0, atol=1e-11)
    xo.assert_allclose(p_exact.y, p_expanded.y, rtol=0, atol=2e-10)
    xo.assert_allclose(p_exact.py, p_expanded.py, rtol=0, atol=1e-11)
    xo.assert_allclose(p_exact.zeta, p_expanded.zeta, rtol=0, atol=1e-12)
    xo.assert_allclose(p_exact.delta, p_expanded.delta, rtol=0, atol=1e-14)
